//Copyright (c) 2014, Jesús Martín Berlanga. All rights reserved. 
//Distributed under the BSD licence. Read "com/jme3/ai/license.txt".
package com.jme3.ai.agents.behaviors.npc.steering;

import com.jme3.ai.agents.Agent;
import com.jme3.math.Vector3f;
import com.jme3.scene.Spatial;

/**
 * Arrival behavior is identical to seek while the character is far from its
 * target. But instead of moving through the target at full speed, this behavior
 * causes the character to slow down as it approaches the target, eventually
 * slowing to a stop coincident with the target.
 *
 * @see SeekBehavior
 *
 * @author Jesús Martín Berlanga
 * @version 1.2.1
 */
public class ArriveBehavior extends SeekBehavior {

    /**
     * Distance of targetPosition that is acceptable.
     */
    private static final float ERROR_FACTOR = 0.001f;
    /**
     * Rate of negative acceleration.
     */
    private float slowingDistance;

    /**
     * The slowingDistance is (0.1 * distance betwen agents) by default.
     *
     * @see SeekBehavior#SeekBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent)
     */
    public ArriveBehavior(Agent agent, Agent target) {
        super(agent, target);
        this.slowingDistance = agent.distanceRelativeToGameEntity(target) * 0.25f;
    }

    /**
     * The slowingDistance is (0.1 * distance betwen agents) by default.
     *
     * @see SeekBehavior#SeekBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent, com.jme3.scene.Spatial)
     */
    public ArriveBehavior(Agent agent, Agent target, Spatial spatial) {
        super(agent, target, spatial);
        this.slowingDistance = agent.distanceRelativeToGameEntity(target) * 0.25f;
    }

    /**
     * The slowingDistance is (0.1 * distance betwen agents) by default.
     *
     * @see SeekBehavior#SeekBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent)
     */
    public ArriveBehavior(Agent agent, Vector3f seekingPosition) {
        super(agent, seekingPosition);
        this.slowingDistance = agent.getLocalTranslation().subtract(seekingPosition).length() * 0.1f;
    }

    /**
     * The slowingDistance is (0.1 * distance betwen agents) by default.
     *
     * @see SeekBehavior#SeekBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent, com.jme3.scene.Spatial)
     */
    public ArriveBehavior(Agent agent, Vector3f seekingPosition, Spatial spatial) {
        super(agent, seekingPosition, spatial);
        this.slowingDistance = agent.getLocalTranslation().subtract(seekingPosition).length() * 0.1f;
    }

    /**
     * @param slowingDistance Distance where the agent will start slowing
     * @throws NegativeSlowingDistanceException If slowingDistance is lower than
     * 0
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent)
     */
    public ArriveBehavior(Agent agent, Agent target, float slowingDistance) {
        super(agent, target);
        this.validateSlowingDistance(slowingDistance);
        this.slowingDistance = slowingDistance;
    }

    /**
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent, float)
     * @see
     * AbstractSteeringBehaviour#AbstractSteeringBehaviour(com.jme3.ai.agents.Agent,
     * com.jme3.scene.Spatial)
     */
    public ArriveBehavior(Agent agent, Agent target, float slowingDistance, Spatial spatial) {
        super(agent, target, spatial);
        this.validateSlowingDistance(slowingDistance);
        this.slowingDistance = slowingDistance;
    }

    /**
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent, float)
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.math.Vector3f)
     */
    public ArriveBehavior(Agent agent, Vector3f seekingPosition, float slowingDistance) {
        super(agent, seekingPosition);
        this.validateSlowingDistance(slowingDistance);
        this.slowingDistance = slowingDistance;
    }

    /**
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.ai.agents.Agent, float)
     * @see ArriveBehavior#ArriveBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.math.Vector3f, com.jme3.scene.Spatial)
     */
    public ArriveBehavior(Agent agent, Vector3f seekingPosition, float slowingDistance, Spatial spatial) {
        super(agent, seekingPosition, spatial);
        this.validateSlowingDistance(slowingDistance);
        this.slowingDistance = slowingDistance;
    }

    private void validateSlowingDistance(float slowingDistance) {
        if (slowingDistance < 0) {
            throw new SteeringExceptions.NegativeValueException("The slowing distance value can not be negative.", slowingDistance);
        }
    }

    /**
     * Calculate steering vector.
     *
     * @return steering vector
     *
     * @see AbstractStrengthSteeringBehaviour#calculateRawSteering()
     * @see SeekBehavior#calculateRawSteering()
     */
    @Override
    protected Vector3f calculateRawSteering() {
        float distanceToTarget;
        float radious = 0;

        if (this.getTarget() != null) {
            distanceToTarget = this.agent.distanceRelativeToGameEntity(this.getTarget());
            radious = this.getTarget().getRadius();
        } else if (this.getSeekingPosition() != null) {
            distanceToTarget = this.agent.getLocalTranslation().subtract(this.getSeekingPosition()).length();
        } else {
            return new Vector3f(); //We dont have any target or location to arrive 
        }
        if (distanceToTarget < radious + ArriveBehavior.ERROR_FACTOR) {
            this.setBrakingFactor(0);
        } else if (distanceToTarget < this.slowingDistance) {
            this.setBrakingFactor(distanceToTarget / this.slowingDistance);
        } else {
            this.setBrakingFactor(1);
        }

        return super.calculateRawSteering();
    }
}
