//Copyright (c) 2014, Jesús Martín Berlanga. All rights reserved. 
//Distributed under the BSD licence. Read "com/jme3/ai/license.txt".
package com.jme3.ai.agents.behaviors.npc.steering;

import com.jme3.ai.agents.Agent;
import com.jme3.ai.agents.Team;
import com.jme3.math.Vector3f;
import com.jme3.scene.Spatial;
import java.util.List;

/**
 * The queuing results from a steering behavior which produces braking
 * (deceleration) when the vehicle detects other vehicles which are: nearby, in
 * front of, and moving slower than itself. <br> <br>
 *
 * In most cases you will combine this behavior with seek, separation and/or
 * avoid. Queuing is designed to be combined with the behaviors mentioned before
 * in order to to leave a large 'room' through a narrow 'doorway': "drawn toward
 * the 'doorway' by seek behavior, avoid walls, and maintain separation from
 * each other."
 *
 * @author Jesús Martín Berlanga
 * @version 1.0.1
 */
public class QueuingBehavior extends AbstractStrengthSteeringBehavior {

    private List<Agent> neighbours;
    private float minDistance;

    /**
     * @param neighbours Queue of agents
     * @param minDistance Min. distance from center to center to consider a
     * neighbour as an obstacle
     *
     * @throws SteeringExceptions.NegativeValueException If minDistance is lower
     * than 0
     *
     * @see
     * AbstractStrengthSteeringBehavior#AbstractStrengthSteeringBehavior(com.jme3.ai.agents.Agent)
     */
    public QueuingBehavior(Agent agent, List<Agent> neighbours, float minDistance) {
        super(agent);
        this.validateMinDistance(minDistance);
        this.neighbours = neighbours;
        this.minDistance = minDistance;
    }

    /**
     * @see QueuingBehavior#QueuingBehavior(com.jme3.ai.agents.Agent,
     * java.util.List, float)
     * @see
     * AbstractStrengthSteeringBehavior#AbstractStrengthSteeringBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.scene.Spatial)
     */
    public QueuingBehavior(Agent agent, List<Agent> neighbours, float minDistance, Spatial spatial) {
        super(agent, spatial);
        this.validateMinDistance(minDistance);
        this.neighbours = neighbours;
        this.minDistance = minDistance;
    }

    private void validateMinDistance(float minDistance) {
        if (minDistance < 0) {
            throw new SteeringExceptions.NegativeValueException("The min distance from an obstacle can not be negative.", minDistance);
        }
    }

    /**
     * @see AbstractStrengthSteeringBehavior#calculateRawSteering()
     */
    @Override
    protected Vector3f calculateRawSteering() {
        Vector3f agentVelocity = this.agent.getVelocity();

        int numberObstaclesFactor = 1;
        float distanceFactor = 1;
        float velocityFactor = 1;

        if (agentVelocity != null && !agentVelocity.equals(Vector3f.ZERO)) {
            for (Agent neighbour : this.neighbours) {
                Vector3f neighVel = neighbour.getVelocity();
                float fordwardness = this.agent.forwardness(neighbour);
                float velDiff;
                float distance;

                if (neighbour != this.agent
                        && (distance = this.agent.distanceRelativeToGameEntity(neighbour)) < this.minDistance
                        && fordwardness > 0
                        && neighVel != null
                        && (velDiff = neighVel.length() - agentVelocity.length()) < 0) {
                    distanceFactor *= distance / this.minDistance;
                    velocityFactor *= -velDiff / this.agent.getMoveSpeed();
                    numberObstaclesFactor++;
                }
            }
        }

        this.setBrakingFactor((distanceFactor + velocityFactor + (1 / numberObstaclesFactor)) / 3);
        return new Vector3f();
    }

    public void setNeighbours(List<Agent> neighbours) {
        this.neighbours = neighbours;
    }

    public void setNeighboursFromTeam(Team team) {
        this.neighbours = team.getMembers();
    }
}