//Copyright (c) 2014, Jesús Martín Berlanga. All rights reserved. 
//Distributed under the BSD licence. Read "com/jme3/ai/license.txt".
package com.jme3.ai.agents.behaviors.npc.steering;

import com.jme3.ai.agents.Agent;
import com.jme3.ai.agents.behaviors.npc.steering.SteeringExceptions.IllegalIntervalException;
import com.jme3.math.Vector3f;
import com.jme3.scene.Spatial;

/**
 * Wander is a type of random steering. This behavior retain steering direction
 * state and make small random displacements to it each frame. Thus at one frame
 * the character may be turning up and to the right, and on the next frame will
 * still be turning in almost the same direction.
 *
 * @see SimpleWanderBehavior
 *
 * @author Jesús Martín Berlanga
 * @version 2.0
 */
public class RelativeWanderBehavior extends SimpleWanderBehavior {

    private float relativeFactor;
    private Vector3f lastSteer;
    
    /**
     * @param relativeFactor How much should differ each new wander force ? A
     * value near to 0 means that each new force must differ slightly from the
     * previous one. 1 means that the behaviour will work as a simple wander
     * behavior
     * @see SimpleWanderBehavior
     * @see SimpleWanderBehavior#SimpleWanderBehavior(com.jme3.ai.agents.Agent,
     * float, float, float) 
     * @throws IllegalIntervalException If relative factor is not contained in 
     * the [0, 1] interval.
     */
    public RelativeWanderBehavior(Agent agent, float rX, float rY, float rZ, float relativeFactor) {
        super(agent, rX, rY, rZ);
        this.validateRelativeFactor(relativeFactor);
        this.relativeFactor = relativeFactor;
        this.currentSteer = this.newRandomSteer();
    }

    /**
     * @see
     * RelativeWanderBehavior#RelativeWanderBehavior(com.jme3.ai.agents.Agent,
     * com.jme3.math.Vector3f, com.jme3.math.Vector3f, float)
     * @see SimpleWanderBehavior#SimpleWanderBehavior(com.jme3.ai.agents.Agent, 
     * float, float, float, com.jme3.scene.Spatial) 
     */
    public RelativeWanderBehavior(Agent agent, float rX, float rY, float rZ, float relativeFactor, Spatial spatial) {
        super(agent, rX, rY, rZ);
        this.validateRelativeFactor(relativeFactor);
        this.relativeFactor = relativeFactor;
        this.currentSteer = this.newRandomSteer();
    }

    private void validateRelativeFactor(float factor) {
        if(factor < 0 || factor > 1)
            throw new IllegalIntervalException("relative", factor);
    }
    
    /**
     * @see SimpleWanderBehavior#changeSteer(float)
     */
    @Override
    protected void changeSteer(float tpf) {
        time -= tpf;
        if (time <= 0) {
            lastSteer = currentSteer;
            Vector3f randomSteer = this.newRandomSteer();
            randomSteer.setX(lastSteer.x*(1-relativeFactor) + randomSteer.getX()*relativeFactor);
            randomSteer.setY(lastSteer.y*(1-relativeFactor) + randomSteer.getY()*relativeFactor);
            randomSteer.setZ(lastSteer.z*(1-relativeFactor) + randomSteer.getZ()*relativeFactor);
            currentSteer = randomSteer;
            time = timeInterval;
        }
    }
        
    /**
     * @param relativeFactor How much should differ each new wander force ? A
     * value near to 0 means that each new force must differ slightly from the
     * previous one. 1 means that the behaviour will work as a simple wander
     * behavior
     * @throws IllegalIntervalException If relative factor is not contained in 
     * the [0, 1] interval.
     */
    public void setRelativeFactor(float factor) {
        this.validateRelativeFactor(factor);
        this.relativeFactor = factor;
    }
    /**
     * @return relativeFactor How much differ each new wander force. A
     * value near to 0 means that each new force differ slightly from the
     * previous one. 1 means that the behaviour works as a simple wander
     * behavior.
     */
    public float getRelativeFactor() {
        return this.relativeFactor;
    }
}