//Copyright (c) 2014, Jesús Martín Berlanga. All rights reserved. 
//Distributed under the BSD licence. Read "com/jme3/ai/license.txt".
package com.jme3.ai.agents.behaviors.npc.steering;

import com.jme3.ai.agents.Agent;
import com.jme3.math.Vector3f;
import com.jme3.scene.Spatial;

/**
 * Purpose of seek behavior is to steer agent towards a specified position or
 * object.<br>
 *
 * You can seek another agent or a specific space location.
 *
 * @author Tihomir Radosavljević
 * @author Jesús Martín Berlanga
 * @version 1.5.1
 */
public class SeekBehavior extends AbstractStrengthSteeringBehavior {

    /**
     * Agent whom we seek.
     */
    private Agent target;
    private Vector3f seekingPosition;

    public SeekBehavior(Agent agent) {
        super(agent);
    }

    /**
     * Constructor for seek behaviour.
     *
     * @param agent to whom behavior belongs
     * @param target agent whom we seek
     */
    public SeekBehavior(Agent agent, Agent target) {
        super(agent);
        this.target = target;
    }

    /**
     * Constructor for seek behavior.
     *
     * @param agent to whom behavior belongs
     * @param seekingPos position that we seek
     * @param spatial active spatial during excecution of behavior
     */
    public SeekBehavior(Agent agent, Agent target, Spatial spatial) {
        super(agent, spatial);
        this.target = target;
    }

    /**
     * Constructor for seek behavior.
     *
     * @param agent to whom behavior belongs
     * @param seekingPosition position that we seek
     */
    public SeekBehavior(Agent agent, Vector3f seekingPosition) {
        super(agent);
        this.seekingPosition = seekingPosition;
    }

    /**
     * Constructor for seek behavior.
     *
     * @param agent to whom behavior belongs
     * @param target agent from we seek
     * @param spatial active spatial during excecution of behavior
     */
    public SeekBehavior(Agent agent, Vector3f seekingPosition, Spatial spatial) {
        super(agent, spatial);
        this.seekingPosition = seekingPosition;
    }

    /**
     * Calculate steering vector.
     *
     * @return steering vector
     *
     * @see AbstractStrengthSteeringBehavior#calculateRawSteering()
     * 
     * @author Jesús Martín Berlanga
     */
    @Override
    protected Vector3f calculateRawSteering() {
        Vector3f desiredVelocity;

        if (this.target != null) {
            desiredVelocity = target.getLocalTranslation().subtract(agent.getLocalTranslation());
        } else if (this.seekingPosition != null) {
            desiredVelocity = this.seekingPosition.subtract(agent.getLocalTranslation()).normalize();
        } else {
            return new Vector3f(); //We do not have a target or position to seek
        }
        Vector3f aVelocity = this.agent.getVelocity();

        if (aVelocity == null) {
            aVelocity = new Vector3f();
        }

        return desiredVelocity.subtract(aVelocity);
    }

    /**
     * Get agent from we seek.
     *
     * @return agent
     */
    public Agent getTarget() {
        return target;
    }

    /**
     * Setting agent from we seek.
     *
     * @param target
     */
    public void setTarget(Agent target) {
        this.target = target;
        this.seekingPosition = null;
    }

    /** @author Jesús Martín Berlanga */
    public Vector3f getSeekingPosition() {
        return this.seekingPosition;
    }

    /** @author Jesús Martín Berlanga */
    public void setSeekingPosition(Vector3f seekingPosition) {
        this.seekingPosition = seekingPosition;
        this.target = null;
    }
}
